Attribute VB_Name = "modLivelli"
'==========================================================
' Descrizione.....: Subroutine CONREC
' Nome dei Files..: Livelli.bas
' Data............: Nov. 1990 - 1 Versione in QuickBasic,
'                   F. Languasco - URL Colworth
' Aggiornamento...: 26/7/1999
' Aggiornamento...: 24/7/2001 (aggiunta etichettatura delle
'                   linee di livello).
' Versione........: 1.1 a 32 bits.
' Sistema.........: Visual Basic 6.0 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'==========================================================
'
'   Subroutine CONREC per il tracciamento delle curve di livello:
'
'   I valori della superficie da tracciare devono essere passati
'   nella matrice D(0 to IUB, 0 to JUB): questa rappresenta una
'   griglia con colonne da 0 a IUB e righe da 0 a JUB.
'
'   Le coordinate reali di colonne e righe, alle quali corrispondono
'   i valori di livello della matrice D(.., ..), vengono passate nei
'   vettori X(0 to IUB), Y(0 to JUB): queste non devono necessariamente
'   essere equispaziate e normalmente, corrispondono alla scalatura
'   del foglio (Page) su cui si vuole disegnare.
'
'   Gli NL valori a cui si intendono tracciare le linee di livello
'   vengono passati nel vettore Z(0 to NL - 1): questi valori
'   non devono necessariamente essere equispaziati ma devono essere
'   in ordine crescente.
'   Nel presente programma deve essere MaxNL <= 15 per avere ogni linea
'   tracciata con un colore diverso di QBColor (16 colori meno il nero).
'   Volendo si potrebbe passare a CONREC un ulteriore vettore
'   (e.g. ZLTipo(0 to NL - 1)) che fornisca, per ogni livello,
'   colore e/o tipo di linea. Modificare in questo caso l' istruzione
'   DRAWIT: della Sub CONREC (ed il Page.ForeColor delle etichette dei
'   livelli).
'
'   Page e' il Picture Box sul quale si vuole disegnare: puo' essere,
'   anziche' un PictureBox, qualsiasi oggetto che supporti il metodo
'   Line.
'
'   Msg$ viene ritornata vuota se CONREC non ha trovato errori,
'   altrimenti contiene la descrizione dell' errore.
'
'   Se fzLbl = True vengono scritti i valori delle linee di livello:
'   in questo caso e' possibile passare anche una stringa di formato,
'   zTxtF$, per la rappresentazione dei valori (default = "0.000E-00").
'
'   La variabile fStop puo' essere impostata a True, nel programma
'   chiamante, per interrompere il tracciamento delle curve di livello.
'
Option Explicit
'
Private Type zLbl_Type
    bLbl As Boolean         ' Flag di livello gia' etichettato.
    x1 As Single            ' Posizioni degli angoli
    x2 As Single            ' dell' area impegnata
    y1 As Single            ' dall' etichetta.
    y2 As Single            '
End Type
Public Sub CONREC(ByVal Page As PictureBox, d!(), X!(), Y!(), z!(), Msg$ _
    , Optional ByVal fzLbl As Boolean = False, Optional ByVal zTxtF$ = "0.000E-00" _
    , Optional ByRef fStop As Boolean = False)
'
'   Subroutine CONREC
'   Da: A countouring subroutine.
'   Di: P. Bourke
'   Byte Jun/1987 - Pg. 143, 150.
'
'   Input Variables to CONREC:
'
'   IUB&, JUB&               Index bounds of the Data Array.
'   D!(0 to IUB&,0 to JUB&)  Matrix for the Data Surface.
'   X!(0 to IUB&)            Data Array for Column coordinates.
'   Y!(0 to JUB&)            Data Array for Row    coordinates.
'   NL&                      Number of Contour Levels.
'   Z!(0 to NL& - 1)         Contour Levels in increasing order.
'
    Dim IUB&, JUB&, NL_1&
    Dim K&, K1&, J&, I&, M&, M1&, M2&, M3&, CAS&, PageDS&
    Dim x1!, y1!, x2!, y2!, DMin!, DMax!
'
    ' Definizioni per le etichette dei livelli:
    Dim PageFN$, PageFS&, PageFC&, PageFT As Boolean
    Dim zTxtW!, zTxtH!, zTxtR!, zTxtB!
'
    Dim h!(0 To 4), XH!(0 To 4), YH!(0 To 4), ISH&(0 To 4)
'
    Dim im&(0 To 3)
    im(0) = 0: im(1) = 1: im(2) = 1:  im(3) = 0
'
    Dim JM&(0 To 3)
    JM(0) = 0: JM(1) = 0: JM(2) = 1: JM(3) = 1
'
    Dim CT&(0 To 2, 0 To 2, 0 To 2)
    CT(0, 0, 0) = 0: CT(0, 0, 1) = 0: CT(0, 0, 2) = 8
    CT(0, 1, 0) = 0: CT(0, 1, 1) = 2: CT(0, 1, 2) = 5
    CT(0, 2, 0) = 7: CT(0, 2, 1) = 6: CT(0, 2, 2) = 9
    CT(1, 0, 0) = 0: CT(1, 0, 1) = 3: CT(1, 0, 2) = 4
    CT(1, 1, 0) = 1: CT(1, 1, 1) = 3: CT(1, 1, 2) = 1
    CT(1, 2, 0) = 4: CT(1, 2, 1) = 3: CT(1, 2, 2) = 0
    CT(2, 0, 0) = 9: CT(2, 0, 1) = 6: CT(2, 0, 2) = 7
    CT(2, 1, 0) = 5: CT(2, 1, 1) = 2: CT(2, 1, 2) = 0
    CT(2, 2, 0) = 8: CT(2, 2, 1) = 0: CT(2, 2, 2) = 0
'
    On Error GoTo CONREC_ERR
'
    IUB = UBound(X)
    JUB = UBound(Y)
    NL_1 = UBound(z)    ' NL - 1
    ReDim zLbl(0 To NL_1) As zLbl_Type
'
    If IUB <= 0 Or JUB <= 0 Or NL_1 <= 0 Then _
        Err.Raise 1001, "CONREC", "Error in Input Parameters"
    For K = 1 To NL_1
        If z(K) <= z(K - 1) Then
            Err.Raise 1001, "CONREC", "Error in Input Parameters"
            Exit For
        End If
    Next K
'
'
    PageDS = Page.DrawStyle
    Page.DrawStyle = vbSolid
    If fzLbl Then   ' Si vogliono scrivere i valori dei livelli.
        ' Salva le impostazioni correnti,
        ' assegna quelle nuove e calcola
        ' i parametri di posizione:
        PageFN$ = Page.FontName
        PageFS = Page.FontSize
        PageFC = Page.ForeColor
        PageFT = Page.FontTransparent
        Page.FontName = "Small Fonts"
        Page.FontSize = 7
        Page.FontTransparent = False
        zTxtW = Page.TextWidth(zTxtF$)
        zTxtH = Abs(Page.TextHeight(zTxtF$))
        zTxtR = X(IUB) - zTxtW
        zTxtB = Y(0) + zTxtH
    End If
'
    For J = 0 To JUB - 1
        For I = 0 To IUB - 1
'
            If d(I, J) < d(I, J + 1) Then
                DMin = d(I, J)
            Else
                DMin = d(I, J + 1)
            End If
            If d(I + 1, J) < DMin Then DMin = d(I + 1, J)
            If d(I + 1, J + 1) < DMin Then DMin = d(I + 1, J + 1)
'
            If d(I, J) > d(I, J + 1) Then
                DMax = d(I, J)
            Else
                DMax = d(I, J + 1)
            End If
            If d(I + 1, J) > DMax Then DMax = d(I + 1, J)
            If d(I + 1, J + 1) > DMax Then DMax = d(I + 1, J + 1)
'
            If DMax < z(0) Or DMin > z(NL_1) Or DMax = DMin Then GoTo NOIBOX
'
            For K = 0 To NL_1
                If z(K) < DMin Or z(K) > DMax Then GoTo NOITRI
'
                For M = 4 To 0 Step -1
                    If M > 0 Then
                        h(M) = d(I + im(M - 1), J + JM(M - 1)) - z(K)
                        XH(M) = X(I + im(M - 1))
                        YH(M) = Y(J + JM(M - 1))
                    End If
                    If M = 0 Then
                        h(0) = (h(1) + h(2) + h(3) + h(4)) / 4!
                        XH(0) = (X(I) + X(I + 1)) / 2!
                        YH(0) = (Y(J) + Y(J + 1)) / 2!
                    End If
                    If h(M) > 0 Then ISH(M) = 2
                    If h(M) < 0 Then ISH(M) = 0
                    If h(M) = 0 Then ISH(M) = 1
                Next M
'
                For M = 1 To 4
                    M1 = M
                    M2 = 0
                    M3 = M + 1
                    If M3 = 5 Then M3 = 1
                    CAS = CT(ISH(M1), ISH(M2), ISH(M3))
                    If CAS = 0 Then GoTo CASE0
                    Select Case CAS
                        Case 1
                        x1 = XH(M1)
                        y1 = YH(M1)
                        x2 = XH(M2)
                        y2 = YH(M2)
'
                        Case 2
                        x1 = XH(M2)
                        y1 = YH(M2)
                        x2 = XH(M3)
                        y2 = YH(M3)
'
                        Case 3
                        x1 = XH(M3)
                        y1 = YH(M3)
                        x2 = XH(M1)
                        y2 = YH(M1)
'
                        Case 4
                        x1 = XH(M1)
                        y1 = YH(M1)
                        x2 = (h(M3) * XH(M2) - h(M2) * XH(M3)) / (h(M3) - h(M2))
                        y2 = (h(M3) * YH(M2) - h(M2) * YH(M3)) / (h(M3) - h(M2))
'
                        Case 5
                        x1 = XH(M2)
                        y1 = YH(M2)
                        x2 = (h(M1) * XH(M3) - h(M3) * XH(M1)) / (h(M1) - h(M3))
                        y2 = (h(M1) * YH(M3) - h(M3) * YH(M1)) / (h(M1) - h(M3))
'
                        Case 6
                        x1 = XH(M3)
                        y1 = YH(M3)
                        x2 = (h(M2) * XH(M1) - h(M1) * XH(M2)) / (h(M2) - h(M1))
                        y2 = (h(M2) * YH(M1) - h(M1) * YH(M2)) / (h(M2) - h(M1))
'
                        Case 7
                        x1 = (h(M2) * XH(M1) - h(M1) * XH(M2)) / (h(M2) - h(M1))
                        y1 = (h(M2) * YH(M1) - h(M1) * YH(M2)) / (h(M2) - h(M1))
                        x2 = (h(M3) * XH(M2) - h(M2) * XH(M3)) / (h(M3) - h(M2))
                        y2 = (h(M3) * YH(M2) - h(M2) * YH(M3)) / (h(M3) - h(M2))
'
                        Case 8
                        x1 = (h(M3) * XH(M2) - h(M2) * XH(M3)) / (h(M3) - h(M2))
                        y1 = (h(M3) * YH(M2) - h(M2) * YH(M3)) / (h(M3) - h(M2))
                        x2 = (h(M1) * XH(M3) - h(M3) * XH(M1)) / (h(M1) - h(M3))
                        y2 = (h(M1) * YH(M3) - h(M3) * YH(M1)) / (h(M1) - h(M3))
'
                        Case 9
                        x1 = (h(M1) * XH(M3) - h(M3) * XH(M1)) / (h(M1) - h(M3))
                        y1 = (h(M1) * YH(M3) - h(M3) * YH(M1)) / (h(M1) - h(M3))
                        x2 = (h(M2) * XH(M1) - h(M1) * XH(M2)) / (h(M2) - h(M1))
                        y2 = (h(M2) * YH(M1) - h(M1) * YH(M2)) / (h(M2) - h(M1))
                    End Select
'
                    ' Gestione di un eventuale Flag di fine ciclo:
                    DoEvents
                    If fStop Then GoTo CONREC_END
'
                    ' Disegna un segmento della linea di livello K:
DRAWIT:             Page.Line (x1, y1)-(x2, y2), QBColor(K + 1)
'
                    ' Etichette delle linee di livello:
                    If fzLbl And (Not zLbl(K).bLbl) Then    ' Prova a scrivere il
                                                            ' valore della linea
                                                            ' di livello K alla
                                                            ' posizione x2, y2.
                        ' Scarta le posizioni fuori quadro:
                        If (zTxtR < x2) Or (y2 < zTxtB) Then GoTo CASE0
                        ' e quelle gia' impegnate:
                        For K1 = 0 To NL_1
                            If (zLbl(K1).x1 < x2) And (x2 < zLbl(K1).x2) _
                            And (zLbl(K1).y1 < y2) And (y2 < zLbl(K1).y2) _
                                Then GoTo CASE0
                        Next K1
'
                        ' La posizione x2, y2 e' OK:
                        zLbl(K).bLbl = True     ' Livello K etichettato.
                        zLbl(K).x1 = x2 - zTxtW ' Area
                        zLbl(K).x2 = x2 + zTxtW ' impegnata
                        zLbl(K).y1 = y2 - zTxtH ' dalla
                        zLbl(K).y2 = y2 + zTxtH ' etichetta K.
'
                        Page.ForeColor = QBColor(K + 1)
                        Page.Print Format$(z(K), zTxtF$)
                    End If
'
CASE0:          Next M
'
NOITRI:     Next K
'
NOIBOX: Next I
'
    Next J
'
CONREC_END:
    Page.DrawStyle = PageDS
    If fzLbl Then
        ' Ripristina le impostazioni:
        Page.FontName = PageFN$
        Page.FontSize = PageFS
        Page.ForeColor = PageFC
        Page.FontTransparent = PageFT
    End If
CONREC_ERR:
    Msg$ = IIf(Err = 0, "", Err.Description)
'
'
'
End Sub

