Attribute VB_Name = "modPoliRad"
'==============================================================
' Descrizione.....: Routines per il calcolo delle radici
'                   dei polinomi.
' Nome dei Files..: PoliRad.bas, modComplex
' Data............: 15/2/2000
' Aggiornamento...: 22/7/2001 (eliminate alcune inconsistenze).
' Versione........: 1.1 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'==============================================================
'   Comprende:
'
'   Quadratica:     Trova le radici, reali e complesse,
'                   di un' equazione di 2 grado (*).
'   QuadraticaSR:   Trova solo le radici reali
'                   di un' equazione di 2 grado (*).
'   DRTEQ3:         Trova le radici, reali e complesse,
'                   di un' equazione di 3 grado (*).
'   Cubica:         Trova le radici, reali e complesse,
'                   di un' equazione di 3 grado (*).
'   CubicaSR:       Trova solo le radici reali
'                   di un' equazione di 3 grado (*).
'   Quartica:       Trova le radici, reali e complesse,
'                   di un' equazione di 4 grado (*).
'   QuarticaSR:     Trova solo le radici reali
'                   di un' equazione di 4 grado (*).
'   POLRT:          Trova tutte le radici reali di un polinomio
'                   di grado > 2 a coefficienti reali.
'   ZRhqr:          Trova tutte le radici, reali e complesse
'                   di un polinomio a coefficienti reali.
'   ZRoots:         Trova tutte le radici, reali e complesse
'                   di un polinomio a coefficienti complessi.
'
'   PoliVal:        Ritorna il valore, reale, di un
'                   polinomio, a coefficienti reali, di
'                   grado N calcolato al punto x (reale).
'
'   CPoliVal:       Ritorna il valore, complesso, di un
'                   polinomio, a coefficienti reali, di
'                   grado N calcolato al punto x (complesso).
'
'   CCPoliVal:      Ritorna il valore, complesso, di un
'                   polinomio, a coefficienti complessi,
'                   di grado N calcolato al punto x (complesso).
'
'   Nota:   Le routines segnate con (*) forniscono soluzioni
'           algebriche (o trigonometriche) delle equazioni e
'           possono introdurre errori di arrotondamento qualora
'           i coefficienti siano molto piccoli o molto grandi.
'           In questi casi e' preferibile usare, dipendentemente
'           dall' applicazione, le routines ZRhqr e ZRoots, il
'           metodo di Jenkins-Traub di RPOLY e CPOLY o la DPRQD
'           della libreria SSP.
'
'   Nota:   Tutti i vettori, Vp(1 To N), contenenti coefficienti
'           di polinomi iniziano dall' indice 1 con Vp(1) = termine
'           costante e Vp(N) = coefficiente della piu' alta potenza
'           di x, i.e:  Vp(I) = coefficiente di x ^ (I - 1))
'
Option Explicit
'
Private Const PI# = 3.14159265358979    ' 4# * Atn(1#)
Public Sub POLRT(cof#(), roots#(), NROOT& _
    , Optional ByVal Tol# = 0.000001, Optional ByVal MTOL& = 100)
'
'   SUBROUTINE POLRT USES NEWTON'S METHOD AND DEFLATION TO FIND THE
'   REAL ROOTS OF A POLYNOMIAL.  THE CALLING PROGRAM MUST SUPPLY THE
'   COEFFICIENTS A(I) (WHERE A(I) IS THE COEFFICIENT OF X ^ (N + 1 - I)),
'   A TOLERANCE TOL AND AN INTEGER MTOL.
'   THE SUBROUTINE RETURNS AN INTEGER NROOT = NUMBER OF REAL ROOTS
'   FOUND AND THE ROOTS (IN ARRAY ROOTS). AT EACH STAGE OF THE
'   DEFLATION, NEWTON'S METHOD TERMINATES WHEN MTOL ITERATIONS HAVE
'   BEEN EXECUTED OR WHEN TWO SUCCESSIVE ITERATES ARE LESS THAN TOL
'   IN ABSOLUTE VALUE.  N MUST BE GREATER THEN 2.
'
'   Nota di traduzione: il vettore cof(), organizzato come di consueto
'                       con cof(1) = termine costante, viene trasformato
'                       in un vettore a() organizzato come richiesto da
'                       questa routine.
'
'   Tradotto dalla Subroutine FORTRAN di:
'       Numerical Analysis
'       di L. W. Johnson, R. D. Riess
'       Addison-Wesley Publishing Company
'
    Dim I&, N&, NP1&, ITR&
    Dim X#, xcorr#, discrm#
'
    N = UBound(cof) - 1
    If N <= 2 Then
        MsgBox "Il grado del polinomio deve essere > 2", vbCritical, " POLROT"
        Exit Sub
    End If
    ReDim a#(1 To N + 1), b#(1 To N + 1), c#(1 To N + 1)
    ReDim roots(1 To N)
'
    ' Aggiusta i coefficienti
    ' nell' ordine richiesto:
    For I = 1 To N + 1
        a(I) = cof(N + 2 - I)
    Next I
'
    NROOT = 0
    NP1 = N + 1
1   ITR = 0
    X = 0#
'
'   SYNTHETIC DIVISION ALGORITHM TO EVALUATE
'   POLYNOMIAL AND ITS DERIVATIVE.
'
2   b(1) = a(1)
    c(1) = b(1)
    For I = 2 To N
        b(I) = X * b(I - 1) + a(I)
        c(I) = X * c(I - 1) + b(I)
    Next I
    b(NP1) = X * b(N) + a(NP1)
'
'   NEWTON'S METHOD UPDATE TO OLD ESTIMATE OF ROOT
'
    xcorr = b(NP1) / c(N)
    X = X - xcorr
    If (Abs(xcorr) < Tol) Then GoTo 4
    ITR = ITR + 1
    If (ITR < MTOL) Then GoTo 2
    Exit Sub
'
4   NROOT = NROOT + 1
    roots(NROOT) = X
'
'   SET UP COEFFICIENTS OF DEFLATED POLYNOMIAL
'
    NP1 = N
    N = N - 1
    For I = 1 To NP1
        a(I) = b(I)
    Next I
'
'   USE QUADRATIC FORMULA WHEN DEFLATED POLYNOMIAL HAS DEGREE 2
'
    If (N > 2) Then GoTo 1
    discrm = a(2) * a(2) - 4# * a(1) * a(3)
    If discrm >= 0# Then
        roots(NROOT + 1) = (-a(2) + Sqr(discrm)) / 2#
        roots(NROOT + 2) = (-a(2) - Sqr(discrm)) / 2#
        NROOT = NROOT + 2
    End If
'
'
'
End Sub

Public Sub ZRoots(a() As Complex, roots() As Complex _
    , Optional ByVal polish As Boolean = True, Optional ByRef Nr&)
'
'   Given the degree M and the complex coefficients a(1 To M + 1) of the polynomial
'   this routine successively calls Laguer and finds all M complex roots.
'   As usual, the first coefficient a(1) is the constant term, while a(M + 1) is the
'   coefficient of the highest power of x.
'   The logical variable polish should be input as True if polishing
'   (also by Laguerre's method) is desired, False if the roots will be
'   subsequently polished by other means.
'
    Dim M&, J&, JJ&, Its&
    Dim X As Complex, b As Complex, c As Complex
    Const eps# = 0.000001   ' A small number.
'
    M = UBound(a) - 1
    ReDim roots(1 To M) 'As Complex
    ReDim ad(1 To M + 1) As Complex
'
    For J = 1 To M + 1      ' Copy of coefficients for successive deflation.
        ad(J) = a(J)
    Next J
'
    For J = M To 1 Step -1  ' Loop over each root to be found.
        X = CCmp(0#, 0#)    ' Start at zero to favor convergence
                            ' to smallest remaining root.
        Call Laguer(ad(), J, X, Its) ' Find the root.
        If Abs(CImm(X)) <= (2# * (eps ^ 2) * Abs(CRea(X))) Then X = CCmp(CRea(X), 0#)
        roots(J) = X
'
        b = ad(J + 1)       ' Forward deflation.
        For JJ = J To 1 Step -1
            c = ad(JJ)
            ad(JJ) = b
            b = CSom(CMol(X, b), c)
        Next JJ
    Next J
'
    If polish Then
        For J = 1 To M      ' Polish the roots using the
                            ' undeflated coefficients.
            Call Laguer(a(), M, roots(J), Its)
        Next J
    End If
'
    Call RootsSort(roots(), Nr)
'
'
'
End Sub

Private Sub Laguer(a() As Complex, ByVal M&, X As Complex, ByRef Its&)
'
'   Given the degree M and the complex coefficients a(1 To M + 1) of the polynomial
'   and given a complex value x, this routine improves x by Laguerre's method
'   until it converges, within the achievable roundoff limit, to a root of the
'   given polynomial. The number of iterations taken is returned as Its.
'   Parameters: EPSS is the estimated fractional roundoff error.
'   We try to break (rare) limit cycles with MR different fractional values,
'   once every MT steps, for MAXIT total allowed iterations.
'
    Const EPSS# = 0.000000002, MR& = 8, MT& = 10, MAXIT& = MT * MR
    Dim Iter&, J&, abx#, abp#, abm#, er#, frac#(1 To MR)
    Dim dx As Complex, x1 As Complex, b As Complex, d As Complex
    Dim f As Complex, g As Complex, h As Complex, sq As Complex
    Dim gp As Complex, gm As Complex, g2 As Complex
'
    frac(1) = 0.5   ' Fractions used to break a limit cycle.
    frac(2) = 0.25
    frac(3) = 0.75
    frac(4) = 0.13
    frac(5) = 0.38
    frac(6) = 0.62
    frac(7) = 0.88
    frac(8) = 1#
'
    For Iter = 1 To MAXIT ' Loop over iterations up to allowed maximum.
        Its = Iter
        b = a(M + 1)
        er = CAbs(b)
        d = CCmp(0#, 0#)
        f = CCmp(0#, 0#)
        abx = CAbs(X)
        For J = M To 1 Step -1  ' Efficient computation of the polynomial
                                ' and its first two derivatives.
            f = CSom(CMol(X, f), d)
            d = CSom(CMol(X, d), b)
            b = CSom(CMol(X, b), a(J))
            er = CAbs(b) + abx * er
        Next J
        er = EPSS * er          ' Estimate of roundoff error in evaluating polynomial.
        If (CAbs(b) <= er) Then ' We are on the root.
            Exit Sub
        Else                    ' The generic case: use Laguerre's formula.
            g = CDiv(d, b)
            g2 = CMol(g, g)
            ' h = g2 - 2# * f / b:
            h = CDif(g2, CDiv(CMol(CCmp(2#, 0#), f), b))
            ' sq = sqrt((M - 1) * (M * h - g2)):
            sq = CSqr(CMol(CCmp(M - 1, 0#), CDif(CMol(CCmp(M, 0#), h), g2)))
            gp = CSom(g, sq)
            gm = CDif(g, sq)
            abp = CAbs(gp)
            abm = CAbs(gm)
            If (abp < abm) Then gp = gm
            If (MAX(abp, abm) > 0#) Then
                dx = CDiv(CCmp(M, 0#), gp)
            Else
                dx = CExp(CCmp(Log(1# + abx), CDbl(Iter)))
            End If
        End If
        x1 = CDif(X, dx)
        If CEgu(X, x1) Then Exit Sub    ' Converged.
'
        If (Iter Mod MT) <> 0 Then
            X = x1
        Else                      ' Every so often we take a fractional step,
                                  ' to break any limit cycle (itself a rare occurrence).
            ' x = x - dx * frac(Iter / MT):
            X = CDif(X, CMol(dx, CCmp(frac(Iter / MT), 0#)))
        End If
    Next Iter
'
    MsgBox "Too many iterations in Laguer"  ' Very unusual.
                                            ' Can occur only for complex roots.
                                            ' Try a different starting guess for the root.
'
'
'
End Sub

Private Function Acos(ByVal X As Double) As Double
'
'   Calcola l' ArcoCoseno di x# con -1 <= x <= 1.
'   Acos e' ritornato in radianti:
'
    If 0# < X And X <= 1# Then
        Acos = Atn(Sqr(1# - X * X) / X)
    ElseIf -1# <= X And X < 0# Then
        Acos = PI + Atn(Sqr(1# - X * X) / X)
    ElseIf X = 0# Then
        Acos = PI / 2#
    Else
        Err.Raise 5, "Acos Function", "Argomento illegale per Acos"
    End If
'
'
'
End Function
Private Function Cur(ByVal X As Double) As Double
'
'   Ritorna la radice cubica del numero reale x:
'
    Cur = CDbl(Sgn(X)) * (Abs(X)) ^ (1# / 3#)
'
'
'
End Function
Public Function Cubica(ByVal a#, ByVal b#, ByVal c#, ByVal d# _
    , x1 As Complex, x2 As Complex, x3 As Complex, Optional ByRef Nr&) As Boolean
'
'   Risolve l' equazione cubica a*x^3 + b*x^2 + c*x + d = 0
'
'    a, b, c, d: coefficienti, reali, dell' equazione.
'    x1, x2, x3: radici, complesse, dell' equazione.
'    Nr:         numero di radici reali.
'
    Dim q1#, q3#, rq2#, R1#, r2#, Theta#
    Dim ar#, br#, b3#, M$
'
    On Error GoTo Cubica_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione cubica.", vbCritical, " Cubica"
        Cubica = False
        Exit Function
    End If
'
    b = b / a
    c = c / a
    d = d / a
'
    b3 = b / 3#
    q1 = (b3) ^ 2 - (c / 3#)
    q3 = q1 * q1 * q1
    R1 = (2# * b ^ 3 - 9# * b * c + 27# * d) / 54#
    r2 = R1 * R1
'
    If r2 < q3 Then     ' Tre radici reali.
        Theta = Acos(R1 / Sqr(q3))
        rq2 = -2# * Sqr(q1)
        x1 = CCmp(rq2 * Cos(Theta / 3#) - b3, 0#)
        x2 = CCmp(rq2 * Cos((Theta + 2# * PI) / 3#) - b3, 0#)
        x3 = CCmp(rq2 * Cos((Theta - 2# * PI) / 3#) - b3, 0#)
        Nr = 3
    Else                ' Una radice reale, le altre due complesse coniugate.
        ar = CDbl(-Sgn(R1)) * Cur(Abs(R1) + Sqr(r2 - q3))
        If ar <> 0# Then br = q1 / ar
        x1 = CCmp((ar + br) - b3, 0#)
        x2 = CCmp(-0.5 * (ar + br) - b3, 0.5 * (ar - br) * Sqr(3#))
        x3 = CCon(x2)
        Nr = 1
    End If
'
'
Cubica_ERR:
    Cubica = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " Cubica"
    End If
'
'
'
End Function

Public Function DRTEQ3(ByVal a#, ByVal b#, ByVal c#, ByVal d# _
    , x1 As Complex, x2 As Complex, x3 As Complex, Optional ByRef Nr&) As Boolean
'
'   Risolve l' equazione cubica a*x^3 + b*x^2 + c*x + d = 0
'
'    a, b, c, d: coefficienti, reali, dell' equazione.
'    x1, x2, x3: radici, complesse, dell' equazione.
'    Nr:         numero di radici reali.
'
    Dim f#, g#, h#
    Dim ar#, br#
    Dim g_2n#, rg_2n#, b3a#
    Dim r#, cr#, t#, ct3#, st3#, M$
'
    On Error GoTo DRTEQ3_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione cubica.", vbCritical, " DRTEQ3"
        DRTEQ3 = False
        Exit Function
    End If
'
    f = ((3# * c / a) - ((b / a) ^ 2)) / 3#
    g = (2# * ((b / a) ^ 3) - (9# * b * c / (a ^ 2)) + (27# * d / a)) / 27#
    h = ((g ^ 2) / 4#) + ((f ^ 3) / 27#)
    g_2n = -g / 2#
    b3a = b / (3# * a)
'
    Select Case h
        Case Is > 0# ' Una radice reale, le altre due complesse coniugate.
        ar = Cur(g_2n + Sqr(h))
        br = Cur(g_2n - Sqr(h))
        x1 = CCmp(ar + br - b3a, 0#)
        x2 = CCmp(-((ar + br) / 2#) - b3a, Sqr(3#) * (ar - br) / 2#)
        x3 = CCon(x2)
        Nr = 1
'
        Case Is = 0# ' Tre radici reali, di cui due coincidenti.
        rg_2n = Cur(g_2n)
        x1 = CCmp(2# * rg_2n - b3a, 0#)
        x2 = CCmp(-rg_2n - b3a, 0#)
        x3 = x2
        Nr = 2
'
        Case Is < 0# ' Tre radici reali distinte.
        r = Sqr((g_2n ^ 2) - h)
        t = Acos(-g / (2# * r))
        cr = Cur(r)
        ct3 = cr * Cos(t / 3#)
        st3 = Sqr(3#) * cr * Sin(t / 3#)
        x1 = CCmp(2# * ct3 - b3a, 0#)
        x2 = CCmp(-ct3 - b3a - st3, 0#)
        x3 = CCmp(-ct3 - b3a + st3, 0#)
        Nr = 3
    End Select
'
'
DRTEQ3_ERR:
    DRTEQ3 = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " DRTEQ3"
    End If
'
'
'
End Function


Public Function QuadraticaSR(ByVal a#, ByVal b#, ByVal c# _
    , ByRef x1#, ByRef x2#, Optional ByRef Nr&) As Boolean
'
'   Trova le radici reali dell' equazione quadratica a*x^2 + b*x + c = 0
'
'    a, b, c:   coefficienti, reali, dell' equazione.
'    x1, x2:    radici, reali, dell' equazione; sono
'               valide solo le prime Nr radici.
'    Nr:        numero di radici reali e distinte.
'
    Dim delta#, q#, mq#, M$
'
    On Error GoTo QuadraticaSR_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione quadratica.", vbCritical, " Quadratica"
        QuadraticaSR = False
        Exit Function
    End If
'
    delta = b ^ 2 - 4# * a * c
    Select Case delta
        Case Is < 0#    ' Nessuna radice reale.
        Nr = 0
'
        Case Is = 0#    ' Due radici reali coincidenti.
        If b = 0 Then
            x1 = 0#
            x2 = x1
        Else
            x1 = -2# * c / b
            x2 = x1
        End If
        Nr = 1
'
        Case Is > 0#    ' Due radici reali distinte.
        q = -0.5 * (b + CDbl(SIGN(1#, b)) * Sqr(delta))
        x1 = q / a
        x2 = c / q
        Nr = 2
    End Select
'
'
QuadraticaSR_ERR:
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " QuadraticaSR"
    End If
'
'
'
End Function
Public Function Quadratica(ByVal a#, ByVal b#, ByVal c# _
    , x1 As Complex, x2 As Complex, Optional ByRef Nr&) As Boolean
'
'   Risolve l' equazione quadratica a*x^2 + b*x + c = 0
'
'    a, b, c:   coefficienti, reali, dell' equazione.
'    x1, x2:    radici, complesse, dell' equazione.
'    Nr:        numero di radici reali e distinte.
'
    Dim delta#, q As Complex, mq#, M$
'
    On Error GoTo Quadratica_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione quadratica.", vbCritical, " Quadratica"
        Quadratica = False
        Exit Function
    End If
'
    delta = b ^ 2 - 4# * a * c
    Select Case delta
        Case Is < 0#    ' Due radici complesse coniugate.
        q = CCmp(-0.5 * b, -0.5 * CDbl(SIGN(1#, b)) * Sqr(-delta))
        x1 = CDiv(q, CCmp(a, 0#))
        x2 = CCon(x1)
        Nr = 0
'
        Case Is = 0#    ' Due radici reali coincidenti.
        If b = 0 Then
            x1 = CCmp(0#, 0#)
            x2 = x1
        Else
            x1 = CCmp(-2# * c / b, 0#)
            x2 = x1
        End If
        Nr = 1
'
        Case Is > 0#    ' Due radici reali distinte.
        q.re = -0.5 * (b + CDbl(SIGN(1#, b)) * Sqr(delta))
        x1 = CCmp(q.re / a, 0#)
        x2 = CCmp(c / q.re, 0#)
        Nr = 2
    End Select
'
'
Quadratica_ERR:
    Quadratica = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " Quadratica"
    End If
'
'
'
End Function



Public Function QuarticaSR(ByVal a#, ByVal b#, ByVal c#, ByVal d#, ByVal e# _
    , ByRef x1#, ByRef x2#, ByRef x3#, ByRef x4#, Optional ByRef Nr&) As Boolean
'
'   Trova le radici reali dell' equazione quartica a*x^4 + b*x^3 + c*x^2 + d*x + e = 0
'
'    a, b, c, d, e:  coefficienti, reali, dell' equazione.
'    x1, x2, x3, x4: radici, reali, dell' equazione; sono
'                    valide solo le prime Nr radici.
'    Nr:             numero di radici reali.
'
    Dim J&, Nc&, p#, q#, r#
    Dim ac#, bc#, cc#, dc#, xc#(1 To 3)
    Dim xK2#, xK#, xL2#, xL#, sqp#, sqm#, cor#, M$
'
    On Error GoTo QuarticaSR_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione quartica.", vbCritical, " Quartica"
        QuarticaSR = False
        Exit Function
    End If
'
    p = (-3# * b ^ 2 + 8# * a * c) / (8# * a ^ 2)
    q = (b ^ 3 - 4# * a * b * c + 8# * d * a ^ 2) / (8# * a ^ 3)
    r = (-3# * b ^ 4 + 16# * a * c * b ^ 2 - 64# * b * d * a ^ 2 _
    + 256# * e * a ^ 3) / (256# * a ^ 4)
    cor = b / (4# * a)
'
'
    ac = 8#
    bc = -4# * p
    cc = -8# * r
    dc = 4# * p * r - q ^ 2
    CubicaSR ac, bc, cc, dc, xc(1), xc(2), xc(3), Nc
'
    For J = 2 To Nc
        If xc(1) < xc(J) Then xc(1) = xc(J)
    Next J
    xK2 = 2# * xc(1) - p
    xK = Sqr(xK2)
'
    If xK = 0# Then
        xL2 = xc(1) ^ 2 - r
        If xL2 < 0# Then
            MsgBox "Nessuna radice."
            Exit Function
        End If
        xL = Sqr(xL2)
    Else
        xL = q / (2# * xK)
    End If
'
    sqp = xK2 - 4# * (xc(1) + xL)
    sqm = xK2 - 4# * (xc(1) - xL)
'
    If (sqp >= 0#) And (sqm >= 0#) Then     ' Quattro radici reali.
        x1 = 0.5 * (xK + Sqr(sqp)) - cor
        x2 = 0.5 * (xK - Sqr(sqp)) - cor
        x3 = 0.5 * (-xK + Sqr(sqm)) - cor
        x4 = 0.5 * (-xK - Sqr(sqm)) - cor
        Nr = 4
    ElseIf (sqp >= 0#) And (sqm < 0#) Then  ' Due radici reali.
        x1 = 0.5 * (xK + Sqr(sqp)) - cor
        x2 = 0.5 * (xK - Sqr(sqp)) - cor
        Nr = 2
    ElseIf (sqp < 0#) And (sqm >= 0#) Then  ' Due radici reali.
        x1 = 0.5 * (-xK + Sqr(sqm)) - cor
        x2 = 0.5 * (-xK - Sqr(sqm)) - cor
        Nr = 2
    ElseIf (sqp < 0#) And (sqm < 0#) Then   ' Nessuna radice reale.
        Nr = 0
    End If
'
'
QuarticaSR_ERR:
    QuarticaSR = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " QuarticaSR"
    End If
'
'
'
End Function
Public Function Quartica(ByVal a#, ByVal b#, ByVal c#, ByVal d#, ByVal e# _
    , x1 As Complex, x2 As Complex, x3 As Complex, x4 As Complex, Optional ByRef Nr&) _
    As Boolean
'
'   Risolve l' equazione quartica a*x^4 + b*x^3 + c*x^2 + d*x + e = 0
'
'    a, b, c, d, e:  coefficienti, reali, dell' equazione.
'    x1, x2, x3, x4: radici, complesse, dell' equazione.
'    Nr:             numero di radici reali.
'
    Dim J&, p#, q#, r#
    Dim ac#, bc#, cc#, dc#, xc#(1 To 3), Nc&
    Dim xK2#, xK#, xL2#, xL#, sqp#, sqm#, cor#, M$
'
    On Error GoTo Quartica_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione quartica.", vbCritical, " Quartica"
        Quartica = False
        Exit Function
    End If
'
    p = (-3# * b ^ 2 + 8# * a * c) / (8# * a ^ 2)
    q = (b ^ 3 - 4# * a * b * c + 8# * d * a ^ 2) / (8# * a ^ 3)
    r = (-3# * b ^ 4 + 16# * a * c * b ^ 2 - 64# * b * d * a ^ 2 _
    + 256# * e * a ^ 3) / (256# * a ^ 4)
    cor = b / (4# * a)
'
'
    ac = 8#
    bc = -4# * p
    cc = -8# * r
    dc = 4# * p * r - q ^ 2
    CubicaSR ac, bc, cc, dc, xc(1), xc(2), xc(3), Nc
'
    For J = 2 To Nc
        If xc(1) < xc(J) Then xc(1) = xc(J)
    Next J
    xK2 = 2# * xc(1) - p
    xK = Sqr(xK2)
'
    If xK = 0# Then
        xL2 = xc(1) ^ 2 - r
        If xL2 < 0# Then
            MsgBox "Nessuna radice."
            Exit Function
        End If
        xL = Sqr(xL2)
    Else
        xL = q / (2# * xK)
    End If
'
    sqp = xK2 - 4# * (xc(1) + xL)
    sqm = xK2 - 4# * (xc(1) - xL)
'
    If (sqp >= 0#) And (sqm >= 0#) Then     ' Quattro radici reali.
        x1 = CCmp(0.5 * (xK + Sqr(sqp)) - cor, 0#)
        x2 = CCmp(0.5 * (xK - Sqr(sqp)) - cor, 0#)
        x3 = CCmp(0.5 * (-xK + Sqr(sqm)) - cor, 0#)
        x4 = CCmp(0.5 * (-xK - Sqr(sqm)) - cor, 0#)
        Nr = 4
    ElseIf (sqp >= 0#) And (sqm < 0#) Then  ' Due radici reali e due complesse coniugate.
        x1 = CCmp(0.5 * (xK + Sqr(sqp)) - cor, 0#)
        x2 = CCmp(0.5 * (xK - Sqr(sqp)) - cor, 0#)
        x3 = CCmp(-0.5 * xK - cor, Sqr(-0.25 * sqm))
        x4 = CCon(x3)
        Nr = 2
    ElseIf (sqp < 0#) And (sqm >= 0#) Then  ' Due radici reali e due complesse coniugate.
        x1 = CCmp(0.5 * (-xK + Sqr(sqm)) - cor, 0#)
        x2 = CCmp(0.5 * (-xK - Sqr(sqm)) - cor, 0#)
        x3 = CCmp(0.5 * xK - cor, Sqr(-0.25 * sqp))
        x4 = CCon(x3)
        Nr = 2
    ElseIf (sqp < 0#) And (sqm < 0#) Then   ' Due coppie di radici complesse coniugate.
        x1 = CCmp(-0.5 * xK - cor, Sqr(-0.25 * sqm))
        x2 = CCon(x1)
        x3 = CCmp(0.5 * xK - cor, Sqr(-0.25 * sqp))
        x4 = CCon(x3)
        Nr = 0
    End If
'
'
Quartica_ERR:
    Quartica = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " Quartica"
    End If
'
'
'
End Function

Private Function MAX(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MAX(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMax = vD(J1)
    For J = J1 + 1 To J2
        If vDMax < vD(J) Then vDMax = vD(J)
    Next J
'
    MAX = vDMax
'
'
'
End Function


Public Function CubicaSR(ByVal a#, ByVal b#, ByVal c#, ByVal d# _
    , x1#, x2#, x3#, Optional ByRef Nr&) As Boolean
'
'   Trova le radici reali dell' equazione cubica a*x^3 + b*x^2 + c*x + d = 0
'
'    a, b, c, d: coefficienti, reali, dell' equazione.
'    x1, x2, x3: radici, reali, dell' equazione; sono
'                valide solo le prime Nr radici.
'    Nr:         numero di radici reali.
'
    Dim J&, w#, p#, q#, dis#, phi#, u#(1 To 3), M$
    Const Third# = 1# / 3#
'
    On Error GoTo CubicaSR_ERR
'
    If a = 0# Then
        MsgBox "Non e' un' equazione cubica.", vbCritical, " Cubica"
        CubicaSR = False
        Exit Function
    End If
'
    w = b / a * Third
    p = (c / a * Third - w ^ 2) ^ 3
    q = -0.5 * (2# * w ^ 3 - (c * w - d) / a)
    dis = q ^ 2 + p
'
    If dis < 0# Then    ' Tre radici reali.
        phi = Acos(MIN(1#, MAX(-1#, q / Sqr(-p))))
        p = 2# * (-p) ^ (0.5 * Third)
        For J = 1 To 3
            u(J) = p * Cos((phi + CDbl(2 * J) * PI) / 3#) - w
        Next J
        x1 = MIN(u(1), u(2), u(3))
        x2 = MAX(MIN(u(1), u(2)), MIN(u(1), u(3)), MIN(u(2), u(3)))
        x3 = MAX(u(1), u(2), u(3))
        Nr = 3
    Else                ' Una radice reale.
        dis = Sqr(dis)
        x1 = Cur(q + dis) + Cur(q - dis) - w
        Nr = 1
    End If
'
'
CubicaSR_ERR:
    CubicaSR = (Err = 0)
    If (Err <> 0) Then
        M$ = Err.Number & ":  " & Err.Description & vbNewLine
        M$ = M$ & "Equazione illegale."
        MsgBox M$, vbCritical, " CubicaSR"
    End If
'
'
'
End Function
Private Function MIN(ParamArray vD() As Variant) As Variant
'
'   Implementa la funzione MIN(D1, D2, ...DN) del FORTRAN:
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMin = vD(J1)
    For J = J1 + 1 To J2
        If vDMin > vD(J) Then vDMin = vD(J)
    Next J
'
    MIN = vDMin
'
'
'
End Function


Public Function PoliVal(ByVal X#, a#()) As Double
'
'   Ritorna il valore, reale, di un polinomio di grado N - 1 calcolato al punto x.
'    x:     punto, reale, al quale calcolare il valore del polinomio.
'    a():   coefficienti, reali, del polinomio.
'           Il numero di coefficienti passato, determina il grado del polinomio.
'           a(1) e' il termine costante, a(N) e' il coefficiente della piu'
'           alta potenza di x.
'
    Dim N&, J&
    Dim pv#
'
    N = UBound(a)
'
    pv = 0#
    For J = N To 1 Step -1
        pv = pv * X + a(J)
    Next J
'
    PoliVal = pv
'
'
'
End Function
Public Function CCPoliVal(X As Complex, a() As Complex) As Complex
'
'   Ritorna il valore, complesso, di un polinomio di grado N - 1 calcolato al punto x.
'    x:     punto, complesso, al quale calcolare il valore del polinomio.
'    a():   coefficienti, complessi, del polinomio.
'           Il numero di coefficienti passato, determina il grado del polinomio.
'           a(1) e' il termine costante, a(N) e' il coefficiente della piu'
'           alta potenza di x.
'
    Dim N&, J&
    Dim pv As Complex
'
    N = UBound(a)
'
    pv = CCmp(0#, 0#)
    For J = N To 1 Step -1
        pv = CSom(CMol(pv, X), a(J))
    Next J
'
    CCPoliVal = pv
'
'
'
End Function

Public Function CPoliVal(X As Complex, a#()) As Complex
'
'   Ritorna il valore, complesso, di un polinomio di grado N - 1 calcolato al punto x.
'    x:     punto, complesso, al quale calcolare il valore del polinomio.
'    a():   coefficienti, reali, del polinomio.
'           Il numero di coefficienti passato, determina il grado del polinomio.
'           a(1) e' il termine costante, a(N) e' il coefficiente della piu'
'           alta potenza di x.
'
    Dim N&, J&
    Dim pv As Complex
'
    N = UBound(a)
'
    pv = CCmp(0#, 0#)
    For J = N To 1 Step -1
        pv = CSom(CMol(pv, X), CCmp(a(J), 0#))
    Next J
'
    CPoliVal = pv
'
'
'
End Function
Public Sub ZRhqr(a#(), roots() As Complex, Optional ByRef Nr&)
'
'   Find all the roots of a polynomial with real coefficients, given
'   the degree M and the coefficients a(1 To M + 1). As usual, the first
'   coefficient a(1) is the constant term, while a(M + 1) is the coefficient
'   of the highest power of x. The method is to construct an upper
'   Hessemberg matrix whose eigenvalues are the desired roots,
'   and then use the routine Balanc and Hqr.
'
    Dim M&, J&, JJ&, xr#, Xi#
    Const MaxM& = 50
'
    M = UBound(a) - 1
    ReDim roots(1 To M) 'As Complex
    ReDim hess#(0 To M, 1 To M), rtr#(1 To M), rti#(1 To M)
'
    If (M > MaxM Or a(M + 1) = 0#) Then
        MsgBox "Bad args in ZRhqr."
        Exit Sub
    End If
'
    For JJ = 1 To M     ' Construct the matrix.
        hess(1, JJ) = -a(M + 1 - JJ) / a(M + 1)
        For J = 2 To M
            hess(J, JJ) = 0#
        Next J
        If (JJ <> M) Then hess(JJ + 1, JJ) = 1#
    Next JJ
'
    Call Balanc(hess()) ' Find its eigenvalues.
    Call Hqr(hess(), rtr(), rti())
'
    For J = 1 To M
        roots(J) = CCmp(rtr(J), rti(J))
    Next J
    Call RootsSort(roots(), Nr)
'
'
'
End Sub

Private Sub Balanc(ByRef a#())
'
'   Given an N by N matrix a() stored in an array, this routine replaces
'   it by a balanced matrix with identical eigenvalues.
'   A symmetric matrix is already balanced and is unaffected
'   by this procedure. The parameter RADIX should be the machine's
'   floating-point radix.
'
    Dim I&, J&, N&, Last&
    Dim c#, f#, g#, r#, s#
    Const RADIX# = 2#, SQRDX# = RADIX * RADIX
'
    N = UBound(a, 1)
'
1
    Last = 1
    For I = 1 To N  ' Calculate row and column norms.
        c = 0#
        r = 0#
        For J = 1 To N
            If (J <> I) Then
                c = c + Abs(a(J, I))
                r = r + Abs(a(I, J))
            End If
        Next J
        If (c <> 0# And r <> 0#) Then ' If both are nonzero,
            g = r / RADIX
            f = 1#
            s = c + r
'
2
            If (c < g) Then ' find the integer power of the machine radix that
                            ' comes closest to balancing the matrix.
                f = f * RADIX
                c = c * SQRDX
                GoTo 2
            End If
            g = r * RADIX
'
3
            If (c > g) Then
                f = f / RADIX
                c = c / SQRDX
                GoTo 3
            End If
'
            If ((c + r) / f < 0.95 * s) Then
                Last = 0
                g = 1# / f
                For J = 1 To N ' Apply similarity transformation.
                    a(I, J) = a(I, J) * g
                Next J
                For J = 1 To N
                    a(J, I) = a(J, I) * f
                Next J
            End If
        End If
    Next I
    If (Last = 0) Then GoTo 1
'
'
'
End Sub

Private Sub Hqr(ByRef a#(), ByRef wr#(), ByRef wi#())
'
'   Finds all eigenvalues of an N by N upper Hessenberg matrix a().
'   On input a() can be exactly as output from Elmhes (11.5);
'   on output it is destroyed.
'   The real and imaginary parts of the eigenvalues are returned
'   in wr and wi, respectively.
'
    Dim I&, Its&, J&, K&, L&, M&, N&, Nn&
    Dim anorm, p#, q#, r#, s#, t#, u#, v#, w#, X#, Y#, z#
'
    N = UBound(a, 1)
'
    anorm = 0#  ' Compute matrix norm for possible use
                ' in locating single small subdiagonal element.
    For I = 1 To N
        For J = MAX(I - 1, 1) To N
            anorm = anorm + Abs(a(I, J))
        Next J
    Next I
'
    Nn = N
    t = 0#                  ' Gets changed only by an exceptional shift.
'
1
        If (Nn >= 1) Then   ' Begin search for next eigenvalue.
        Its = 0
'
2
        For L = Nn To 2 Step -1 ' Begin iteration: look for single small
                                ' sub-diagonal element.
            s = Abs(a(L - 1, L - 1)) + Abs(a(L, L))
            If (s = 0#) Then s = anorm
            If (Abs(a(L, L - 1)) + s = s) Then GoTo 3
        Next L
        L = 1
'
3
        X = a(Nn, Nn)
        If (L = Nn) Then            ' One root found.
            wr(Nn) = X + t
            wi(Nn) = 0#
            Nn = Nn - 1
        Else
            Y = a(Nn - 1, Nn - 1)
            w = a(Nn, Nn - 1) * a(Nn - 1, Nn)
            If (L = Nn - 1) Then    ' Two roots found...
                p = 0.5 * (Y - X)
                q = p ^ 2 + w
                z = Sqr(Abs(q))
                X = X + t
                If (q >= 0#) Then   '...a real pair.
                    z = p + SIGN(z, p)
                    wr(Nn) = X + z
                    wr(Nn - 1) = wr(Nn)
                    If (z <> 0#) Then wr(Nn) = X - w / z
                    wi(Nn) = 0#
                    wi(Nn - 1) = 0#
                Else                ' ...a complex pair.
                    wr(Nn) = X + p
                    wr(Nn - 1) = wr(Nn)
                    wi(Nn) = z
                    wi(Nn - 1) = -z
                End If
                Nn = Nn - 2
            Else                    ' No roots found. Continue iteration.
                If (Its = 30) Then MsgBox "Too many iterations in hqr."
                If (Its = 10 Or Its = 20) Then ' Form exceptional shift.
                    t = t + X
                    For I = 1 To Nn
                        a(I, I) = a(I, I) - X
                    Next I
                    s = Abs(a(Nn, Nn - 1)) + Abs(a(Nn - 1, Nn - 2))
                    X = 0.75 * s
                    Y = X
                    w = -0.4375 * s ^ 2
                End If
                Its = Its + 1
                For M = Nn - 2 To L Step -1 ' Form shift and then look for 2 consecu-
                                            ' tive small subdiagonal elements.
                    z = a(M, M)
                    r = X - z
                    s = Y - z
                    p = (r * s - w) / a(M + 1, M) + a(M, M + 1) ' Equation (11.6.23).
                    q = a(M + 1, M + 1) - z - r - s
                    r = a(M + 2, M + 1)
                    s = Abs(p) + Abs(q) + Abs(r) ' Scale to prevent over ow or under ow.
                    p = p / s
                    q = q / s
                    r = r / s
                    If (M = L) Then GoTo 4
                    u = Abs(a(M, M - 1)) * (Abs(q) + Abs(r))
                    v = Abs(p) * (Abs(a(M - 1, M - 1)) + Abs(z) + Abs(a(M + 1, M + 1)))
                    If (u + v = v) Then GoTo 4 ' Equation (11.6.26).
                Next M
'
4
                For I = M + 2 To Nn
                    a(I, I - 2) = 0#
                    If (I <> M + 2) Then a(I, I - 3) = 0#
                Next I
                For K = M To Nn - 1     ' Double QR step on rows l to nn and
                                        ' columns m to nn.
                    If (K <> M) Then
                        p = a(K, K - 1) ' Begin setup of Householder vector.
                        q = a(K + 1, K - 1)
                        r = 0#
                        If (K <> Nn - 1) Then r = a(K + 2, K - 1)
                        X = Abs(p) + Abs(q) + Abs(r)
                        If (X <> 0#) Then
                            p = p / X   ' Scale to prevent over ow or under ow.
                            q = q / X
                            r = r / X
                        End If
                    End If
                    s = SIGN(Sqr(p ^ 2 + q ^ 2 + r ^ 2), p)
                    If (s <> 0#) Then
                        If (K = M) Then
                            If (L <> M) Then a(K, K - 1) = -a(K, K - 1)
                        Else
                            a(K, K - 1) = -s * X
                        End If
                        p = p + s       ' Equations (11.6.24).
                        X = p / s
                        Y = q / s
                        z = r / s
                        q = q / p
                        r = r / p
                        For J = K To Nn ' Row modification.
                            p = a(K, J) + q * a(K + 1, J)
                            If (K <> Nn - 1) Then
                                p = p + r * a(K + 2, J)
                                a(K + 2, J) = a(K + 2, J) - p * z
                            End If
                            a(K + 1, J) = a(K + 1, J) - p * Y
                            a(K, J) = a(K, J) - p * X
                        Next J
                        For I = L To MIN(Nn, K + 3) ' Column modification.
                            p = X * a(I, K) + Y * a(I, K + 1)
                            If (K <> Nn - 1) Then
                                p = p + z * a(I, K + 2)
                                a(I, K + 2) = a(I, K + 2) - p * r
                            End If
                            a(I, K + 1) = a(I, K + 1) - p * q
                            a(I, K) = a(I, K) - p
                        Next I
                    End If
                Next K
                GoTo 2      ' ...for next iteration on current eigenvalue.
            End If
        End If
        GoTo 1              ' ...for next eigenvalue.
    End If
'
'
'
End Sub

Private Sub RootsSort(ByRef roots() As Complex, Optional ByRef Nr&)
'
'   Sort roots by their real parts by straight insertion.
'   Nr: number of real roots.
'
    Dim I&, J&, M&, X As Complex
'
    M = UBound(roots)
'
    For J = 2 To M
        X = roots(J)
        For I = J - 1 To 1 Step -1
            If (CRea(roots(I)) <= CRea(X)) Then GoTo 10
            roots(I + 1) = roots(I)
        Next I
        I = 0
'
10
        roots(I + 1) = X
    Next J
'
    Nr = 0
    For J = 1 To M
        If CImm(roots(J)) = 0# Then Nr = Nr + 1
    Next J
'
'
'
End Sub


Private Function SIGN(ByVal dV#, ByVal DS#) As Double
'
'   Ritorna il valore assoluto di v con il segno di s.
'   Implementa la funzione SIGN del FORTRAN:
'
    If DS < 0 Then
        SIGN = -Abs(dV)
    Else
        SIGN = Abs(dV)
    End If
'
'
'
End Function

