Attribute VB_Name = "modCPOLY"
'========================================================
' Descrizione.....: Routine per il calcolo delle radici,
'                   complesse, di polinomi a coefficienti
'                   complessi.
'                   Implementa il metodo Jenkins-Traub
'                   evitando l' uso di operazioni con
'                   numeri complessi.
' Nome dei Files..: modCPOLY.bas
' Data............: 15/2/2000
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://members.xoom.it/flanguasco/
'                   http://www.mclink.it/personal/MC7061/
'========================================================
'
'   Routine di ingresso: CPOLY
'    CPOLY opr(),opi(), zeror(), zeroi()
'     opr():    vettore della parte reale dei coefficienti del polinomio.
'     opi():    vettore della parte immaginaria dei coefficienti del polinomio.
'               Iniziano dall' indice 1 e la routine CPOLY e'
'               stata modificata per accettare opx(1) come
'               termine costante e opx(N) = coefficiente della piu'
'               alta potenza di x, i.e: opx(I) = coefficiente di x ^ (I - 1))
'     zeror():  vettore delle parti reali delle radici.
'     zeroi():  vettore delle parti immaginarie delle radici.
'
'   Tratta dal programma FORTRAN (vedi file J&T_CC.txt):
'    Title:  CPOLY
'    For:    Zeros of complex polynomial
'    Alg:    Jenkins and Traub
'    Ref:    Comm ACM 15 (1972) 97-99
'
'   I commenti in tutte maiuscole sono originali.
'
Option Explicit
'
Private Nn&
Private pre#(), pim#(), hr#(), hi#()
Private qpr#(), qpi#(), qhr#(), qhi#(), shr#(), shi#()
Private SR#, si#, tr#, ti#, pvr#, pvi#
Private ARE#, MRE# ', INFINY#, ETA#
'
' Vedi MCON:
Private Const vbT& = 15
Private Const vbM& = 308
Private Const vbN& = -323
Private Const BASE# = 10#
Private Const ETA# = BASE ^ (1 - vbT)
Private Const INFINY# = BASE * (1# - BASE ^ (-vbT)) * BASE ^ (vbM - 1)
Private Const SMALNO# = (BASE ^ (vbN + 3)) / BASE ^ 3

Public Sub CPOLY(opr#(), opi#(), ByRef zeror#(), ByRef zeroi#() _
    , Optional ByRef Fail As Boolean)
'
'   FINDS THE ZEROS OF A COMPLEX POLYNOMIAL.
'   OPR, OPI:       DOUBLE PRECISION VECTORS OF REAL AND
'                   IMAGINARY PARTS OF THE COEFFICIENTS IN
'                   ORDER OF increasing POWERS.
'                   As usual, the first coefficient OPx(1) is
'                   the constant term, while OPx(DEGREE + 1) is
'                   the coefficient of the highest power of x.
'   DEGREE:         INTEGER DEGREE OF POLYNOMIAL.
'   ZEROR, ZEROI:   OUTPUT DOUBLE PRECISION VECTORS OF
'                   REAL AND IMAGINARY PARTS OF THE ZEROS.
'   FAIL:           OUTPUT LOGICAL PARAMETER,  TRUE  ONLY IF
'                   LEADING COEFFICIENT IS ZERO OR IF CPOLY
'                   HAS FOUND FEWER THAN DEGREE ZEROS.
'   THE PROGRAM HAS BEEN WRITTEN TO REDUCE THE CHANCE OF OVERFLOW
'   OCCURRING. IF IT DOES OCCUR, THERE IS STILL A POSSIBILITY THAT
'   THE ZEROFINDER WILL WORK PROVIDED THE OVERFLOWED QUANTITY IS
'   REPLACED BY A LARGE NUMBER.
'
    Dim DEGREE&, I&, CNT1&, CNT2&, IDNN2&
    Dim xx#, xxx#, yy#, zr#, zi#, bnd#
    Dim SMALNO#, BASE#
    Dim CONV As Boolean
'
'   Legge il grado del polinomio:
    DEGREE = UBound(opr) - 1
'
    ReDim zeror#(1 To DEGREE), zeroi#(1 To DEGREE)
    ReDim pre#(1 To DEGREE + 1), pim#(1 To DEGREE + 1), hr#(1 To DEGREE + 1)
    ReDim hi#(1 To DEGREE + 1), qpr#(1 To DEGREE + 1), qpi#(1 To DEGREE + 1)
    ReDim qhr#(1 To DEGREE + 1), qhi#(1 To DEGREE + 1), shr#(1 To DEGREE + 1)
    ReDim shi#(1 To DEGREE + 1)
'
'   INITIALIZATION OF CONSTANTS.
'   ' Dichiarate come Private Const:
    'Call MCON(ETA, INFINY, SMALNO, BASE)
    ARE = ETA
    MRE = 2# * Sqr(2#) * ETA
'
    Const COSR# = -6.97564737441253E-02 '-0.060756474
    Const SINR# = 0.997564050259824     ' 0.99756405
    xx = Sin(Atn(1#))                   ' 0.70710678
    yy = -xx
'
    Fail = False
    Nn = DEGREE + 1
'
'   MAKE A COPY OF THE COEFFICIENTS.
'   Inverte l' ordine dei coefficienti
'   per adeguarsi alla struttura originale
'   di CPOLY:
    For I = 1 To DEGREE + 1
        pre(I) = opr(DEGREE + 2 - I)
        pim(I) = opi(DEGREE + 2 - I)
        shr(I) = CMOD(pre(I), pim(I))
    Next I
'
'   ALGORITHM FAILS IF THE LEADING COEFFICIENT IS ZERO.
    If (pre(1) <> 0# Or pim(1) <> 0#) Then GoTo 10
    Fail = True
    Exit Sub
'
'   REMOVE THE ZEROS AT THE ORIGIN IF ANY.
10
    If (pre(Nn) <> 0# Or pim(Nn) <> 0#) Then GoTo 20
    IDNN2 = DEGREE - Nn + 2
    zeror(IDNN2) = 0#
    zeroi(IDNN2) = 0#
    Nn = Nn - 1
    GoTo 10
'
'   SCALE THE POLYNOMIAL.
20
    bnd = SFACTOR(Nn, shr(), SMALNO, BASE)
    If (bnd = 1#) Then GoTo 40
    For I = 1 To Nn
        pre(I) = bnd * pre(I)
        pim(I) = bnd * pim(I)
    Next I
'
'   START THE ALGORITHM FOR ONE ZERO.
40
    If (Nn > 2) Then GoTo 50
'
'   CALCULATE THE FINAL ZERO AND RETURN.
    Call CDIVID(-pre(2), -pim(2), pre(1), pim(1), zeror(DEGREE), zeroi(DEGREE))
    Exit Sub
'
'   CALCULATE BND, A LOWER BOUND ON THE MODULUS OF THE ZEROS.
50
    For I = 1 To Nn
        shr(I) = CMOD(pre(I), pim(I))
    Next I
    bnd = CAUCHY(Nn, shr(), shi())
'
'   OUTER LOOP TO CONTROL 2 MAJOR PASSES WITH DIFFERENT SEQUENCES
'   OF SHIFTS.
    For CNT1 = 1 To 2
'
'   FIRST STAGE CALCULATION, NO SHIFT.
        Call NOSHFT(5)
'
'   INNER LOOP TO SELECT A SHIFT.
        For CNT2 = 1 To 9
'
'   SHIFT IS CHOSEN WITH MODULUS BND AND AMPLITUDE ROTATED BY
'   94 DEGREES FROM THE PREVIOUS SHIFT.
            xxx = COSR * xx - SINR * yy
            yy = SINR * xx + COSR * yy
            xx = xxx
            SR = bnd * xx
            si = bnd * yy
'
'   SECOND STAGE CALCULATION, FIXED SHIFT.
            Call FXSHFT(10 * CNT2, zr, zi, CONV)
            If (Not CONV) Then GoTo 80
'
'   THE SECOND STAGE JUMPS DIRECTLY TO THE THIRD STAGE ITERATION.
'   IF SUCCESSFUL THE ZERO IS STORED AND THE POLYNOMIAL DEFLATED.
            IDNN2 = DEGREE - Nn + 2
            zeror(IDNN2) = zr
            zeroi(IDNN2) = zi
            Nn = Nn - 1
            For I = 1 To Nn
                pre(I) = qpr(I)
                pim(I) = qpi(I)
            Next I
            GoTo 40
'
'   IF THE ITERATION IS UNSUCCESSFUL ANOTHER SHIFT IS CHOSEN.
80
        Next CNT2
'
'   IF 9 SHIFTS FAIL, THE OUTER LOOP IS REPEATED WITH ANOTHER
'   SEQUENCE OF SHIFTS.
    Next CNT1 '100 CONTINUE
'
'   THE ZEROFINDER HAS FAILED ON TWO MAJOR PASSES.
'   RETURN EMPTY HANDED.
    Fail = True
'
'
'
End Sub

Private Sub NOSHFT(ByVal L1&)
'
'   COMPUTES  THE DERIVATIVE  POLYNOMIAL AS THE INITIAL H
'   POLYNOMIAL AND COMPUTES L1 NO-SHIFT H POLYNOMIALS.
'
    Dim I&, J&, JJ&, N&, NM1&
    Dim xni#, t1#, t2#
'
    N = Nn - 1
    NM1 = N - 1
'
    For I = 1 To N
        xni = Nn - I
        hr(I) = xni * pre(I) / CDbl(N)
        hi(I) = xni * pim(I) / CDbl(N)
    Next I
'
    For JJ = 1 To L1
        If (CMOD(hr(N), hi(N)) <= ETA * 10# * CMOD(pre(N), pim(N))) Then GoTo 30
        Call CDIVID(-pre(Nn), -pim(Nn), hr(N), hi(N), tr, ti)
        For I = 1 To NM1
            J = Nn - I
            t1 = hr(J - 1)
            t2 = hi(J - 1)
            hr(J) = tr * t1 - ti * t2 + pre(J)
            hi(J) = tr * t2 + ti * t1 + pim(J)
        Next I
        hr(1) = pre(1)
        hi(1) = pim(1)
        GoTo 50
'
'   IF THE CONSTANT TERM IS ESSENTIALLY ZERO, SHIFT H COEFFICIENTS.
30
        For I = 1 To NM1
            J = Nn - I
            hr(J) = hr(J - 1)
            hi(J) = hi(J - 1)
        Next I
        hr(1) = 0#
        hi(1) = 0#
'
50
    Next JJ
'
'
'
End Sub

Private Sub FXSHFT(ByVal L2&, ByRef zr#, ByRef zi#, ByRef CONV As Boolean)
'
'   COMPUTES L2 FIXED-SHIFT H POLYNOMIALS AND TESTS FOR
'   CONVERGENCE.
'   INITIATES A VARIABLE-SHIFT ITERATION AND RETURNS WITH THE
'   APPROXIMATE ZERO IF SUCCESSFUL.
'   L2:     LIMIT OF FIXED SHIFT STEPS
'   ZR,ZI:  APPROXIMATE ZERO IF CONV IS .TRUE.
'   CONV:   LOGICAL INDICATING CONVERGENCE OF STAGE 3 ITERATION
'
    Dim I&, J&, N&
    Dim otr#, oti#, svsr#, svsi#
    Dim TEST As Boolean, PASD As Boolean, BOOL As Boolean
'
    N = Nn - 1
'
'   EVALUATE P AT S.
    Call POLYEV(Nn, SR, si, pre(), pim(), qpr(), qpi(), pvr, pvi)
    TEST = True
    PASD = False
'
'   CALCULATE FIRST T = -P(S)/H(S).
    Call CALCT(BOOL)
'
'   MAIN LOOP FOR ONE SECOND STAGE STEP.
    For J = 1 To L2
        otr = tr
        oti = ti
'
'   COMPUTE NEXT H POLYNOMIAL AND NEW T.
        Call NEXTH(BOOL)
        Call CALCT(BOOL)
        zr = SR + tr
        zi = si + ti
'
'   TEST FOR CONVERGENCE UNLESS STAGE 3 HAS FAILED ONCE OR THIS
'   IS THE LAST H POLYNOMIAL .
        If (BOOL Or Not TEST Or J = L2) Then GoTo 50
        If (CMOD(tr - otr, ti - oti) >= 0.5 * CMOD(zr, zi)) Then GoTo 40
        If (Not PASD) Then GoTo 30
'
'   THE WEAK CONVERGENCE TEST HAS BEEN PASSED TWICE, START THE
'   THIRD STAGE ITERATION, AFTER SAVING THE CURRENT H POLYNOMIAL
'   AND SHIFT.
        For I = 1 To N
            shr(I) = hr(I)
            shi(I) = hi(I)
        Next I
        svsr = SR
        svsi = si
        Call VRSHFT(10&, zr, zi, CONV)
        If (CONV) Then Exit Sub
'
'   THE ITERATION FAILED TO CONVERGE. TURN OFF TESTING AND RESTORE
'   H,S,PV AND T.
        TEST = False
        For I = 1 To N
            hr(I) = shr(I)
            hi(I) = shi(I)
        Next I
        SR = svsr
        si = svsi
        Call POLYEV(Nn, SR, si, pre(), pim(), qpr(), qpi(), pvr, pvi)
        Call CALCT(BOOL)
        GoTo 50
'
30
        PASD = True
        GoTo 50
'
40
        PASD = False
'
50
    Next J
'
'   ATTEMPT AN ITERATION WITH FINAL H POLYNOMIAL FROM SECOND STAGE.
    Call VRSHFT(10&, zr, zi, CONV)
'
'
'
End Sub

Private Sub VRSHFT(ByVal L3&, ByRef zr#, ByRef zi#, ByRef CONV As Boolean)
'
'   CARRIES OUT THE THIRD STAGE ITERATION.
'   L3:     LIMIT OF STEPS IN STAGE 3.
'   ZR,ZI:  ON ENTRY CONTAINS THE INITIAL ITERATE, IF THE
'           ITERATION CONVERGES IT CONTAINS THE FINAL ITERATE
'           ON EXIT.
'   CONV:   .TRUE. IF ITERATION CONVERGES
'
    Dim I&, J&
    Dim mp#, MS#, omp#, relstp#, R1#, r2#, tp#
    Dim BO As Boolean, BOOL As Boolean
'
    CONV = False
    BO = False
    SR = zr
    si = zi
'
'   MAIN LOOP FOR STAGE THREE.
    For I = 1 To L3
'
'   EVALUATE P AT S AND TEST FOR CONVERGENCE.
        Call POLYEV(Nn, SR, si, pre(), pim(), qpr(), qpi(), pvr, pvi)
        mp = CMOD(pvr, pvi)
        MS = CMOD(SR, si)
        If (mp > 20# * ERREV(Nn, qpr(), qpi(), MS, mp)) Then GoTo 10
'
'   POLYNOMIAL VALUE IS SMALLER IN VALUE THAN A BOUND ON THE ERROR
'   IN EVALUATING P, TERMINATE THE ITERATION.
        CONV = True
        zr = SR
        zi = si
        Exit Sub
'
10
        If (I = 1) Then GoTo 40
        If (BO Or mp < omp Or relstp >= 0.05) Then GoTo 30
'
'   ITERATION HAS STALLED. PROBABLY A CLUSTER OF ZEROS. DO 5 FIXED
'   SHIFT STEPS INTO THE CLUSTER TO FORCE ONE ZERO TO DOMINATE.
        tp = relstp
        BO = True
        If (relstp < ETA) Then tp = ETA
        R1 = Sqr(tp)
        r2 = SR * (1# + R1) - si * R1
        si = SR * R1 + si * (1# + R1)
        SR = r2
        Call POLYEV(Nn, SR, si, pre(), pim(), qpr(), qpi(), pvr, pvi)
        For J = 1 To 5
            Call CALCT(BOOL)
            Call NEXTH(BOOL)
        Next J
        omp = INFINY
        GoTo 50
'
'   EXIT IF POLYNOMIAL VALUE INCREASES SIGNIFICANTLY.
30
        If (mp * 0.1 > omp) Then Exit Sub
'
40
        omp = mp
'
'   CALCULATE NEXT ITERATE.
50
        Call CALCT(BOOL)
        Call NEXTH(BOOL)
        Call CALCT(BOOL)
        If (BOOL) Then GoTo 60
        relstp = CMOD(tr, ti) / CMOD(SR, si)
        SR = SR + tr
        si = si + ti
'
60
    Next I
'
'
'
End Sub

Private Sub CALCT(BOOL As Boolean)
'
'   COMPUTES  T = -P(S)/H(S).
'   BOOL:   LOGICAL, SET TRUE IF H(S) IS ESSENTIALLY ZERO.
'
    Dim N&, hvr#, hvi#
'
    N = Nn - 1
'
'   EVALUATE H(S).
    Call POLYEV(N, SR, si, hr(), hi(), qhr(), qhi(), hvr, hvi)
    BOOL = CMOD(hvr, hvi) < ARE * 10# * CMOD(hr(N), hi(N))
    If (BOOL) Then GoTo 10
    Call CDIVID(-pvr, -pvi, hvr, hvi, tr, ti)
    Exit Sub
'
10
    tr = 0#
    ti = 0#
'
'
'
End Sub

Private Sub NEXTH(ByVal BOOL As Boolean)
'
'   CALCULATES THE NEXT SHIFTED H POLYNOMIAL.
'   BOOL:   LOGICAL, IF .TRUE. H(S) IS ESSENTIALLY ZERO.
'
    Dim J&, N&, NM1&, t1#, t2#
'
    N = Nn - 1
    NM1 = N - 1
'
    If BOOL Then GoTo 20
    For J = 2 To N
        t1 = qhr(J - 1)
        t2 = qhi(J - 1)
        hr(J) = tr * t1 - ti * t2 + qpr(J)
        hi(J) = tr * t2 + ti * t1 + qpi(J)
    Next J
    hr(1) = qpr(1)
    hi(1) = qpi(1)
    Exit Sub
'
'   IF H(S) IS ZERO REPLACE H WITH QH.
20
    For J = 2 To N
        hr(J) = qhr(J - 1)
        hi(J) = qhi(J - 1)
    Next J
    hr(1) = 0#
    hi(1) = 0#
'
'
'
End Sub

Private Sub POLYEV(ByVal PNn&, ByVal psr#, ByVal psi#, ppr#(), ppi#() _
    , ByRef pqr#(), ByRef pqi#(), ByRef ppvr#, ByRef ppvi#)
'
'   EVALUATES A POLYNOMIAL  P  AT  S  BY THE HORNER RECURRENCE
'   PLACING THE PARTIAL SUMS IN Q AND THE COMPUTED VALUE IN PV.
'
    Dim I&, t#
'
    pqr(1) = ppr(1)
    pqi(1) = ppi(1)
    ppvr = pqr(1)
    ppvi = pqi(1)
    For I = 2 To PNn
        t = ppvr * psr - ppvi * psi + ppr(I)
        ppvi = ppvr * psi + ppvi * psr + ppi(I)
        ppvr = t
        pqr(I) = ppvr
        pqi(I) = ppvi
    Next I
'
'
'
End Sub

Private Function ERREV(ByVal ENn&, qr#(), qi#(), ByVal MS#, ByVal mp#) As Double
'
'   BOUNDS THE ERROR IN EVALUATING THE POLYNOMIAL BY THE HORNER
'   RECURRENCE.
'   QR,QI:      THE PARTIAL SUMS
'   MS:         MODULUS OF THE POINT
'   MP:         MODULUS OF POLYNOMIAL VALUE
'   ARE, MRE:   ERROR BOUNDS ON COMPLEX ADDITION AND MULTIPLICATION
'               DOUBLE PRECISION QR(NN),QI(NN),MS,MP,ARE,MRE,E,CMOD
'
    Dim I&, e#
'
    e = CMOD(qr(1), qi(1)) * MRE / (ARE + MRE)
    For I = 1 To ENn
        e = e * MS + CMOD(qr(I), qi(I))
    Next I
'
    ERREV = e * (ARE + MRE) - mp * MRE
'
'
'
End Function

Private Function CAUCHY(ByVal CNn&, ByRef pt#(), ByRef q#()) As Double
'
'   CAUCHY COMPUTES A LOWER BOUND ON THE MODULI OF THE ZEROS OF A
'   POLYNOMIAL - PT IS THE MODULUS OF THE COEFFICIENTS.
'
    Dim I&, N&
    Dim X#, xm#, f#, dx#, df#
'
    pt(CNn) = -pt(CNn)
'
'   COMPUTE UPPER ESTIMATE OF BOUND.
    N = CNn - 1
    X = Exp((Log(-pt(CNn)) - Log(pt(1))) / CDbl(N))
    If (pt(N) = 0#) Then GoTo 20
'
'   IF NEWTON STEP AT THE ORIGIN IS BETTER, USE IT.
    xm = -pt(CNn) / pt(N)
    If (xm < X) Then X = xm
'
'   CHOP THE INTERVAL (0,X) UNTIL F .LE. 0.
20
    xm = X * 0.1
    f = pt(1)
    For I = 2 To CNn
          f = f * xm + pt(I)
    Next I
    If (f <= 0#) Then GoTo 40
    X = xm
    GoTo 20
'
40
    dx = X
'
'   DO NEWTON ITERATION UNTIL X CONVERGES TO TWO DECIMAL PLACES.
50
    If (Abs(dx / X) <= 0.005) Then GoTo 70
'
    q(1) = pt(1)
    For I = 2 To CNn
        q(I) = q(I - 1) * X + pt(I)
    Next I
'
    f = q(CNn)
    df = q(1)
    For I = 2 To N
        df = df * X + q(I)
    Next I
    dx = f / df
    X = X - dx
    GoTo 50
'
70
    CAUCHY = X
'
'
'
End Function

Private Function SFACTOR(ByVal Nn&, pt#(), ByVal SMALNO#, ByVal BASE#) As Double
'
'   RETURNS A SCALE FACTOR TO MULTIPLY THE COEFFICIENTS OF THE
'   POLYNOMIAL. THE SCALING IS DONE TO AVOID OVERFLOW AND TO AVOID
'   UNDETECTED UNDERFLOW INTERFERING WITH THE CONVERGENCE
'   CRITERION.  THE FACTOR IS A POWER OF THE BASE.
'   PT:                     MODULUS OF COEFFICIENTS OF P.
'   ETA,INFINY,SMALNO,BASE: CONSTANTS DESCRIBING THE
'                           FLOATING POINT ARITHMETIC.
'
    Dim I&, X#, hi#, LO#, MAX#, MIN#, SC#, L#
'
'   FIND LARGEST AND SMALLEST MODULI OF COEFFICIENTS.
    hi = Sqr(INFINY)
    LO = SMALNO / ETA
    MAX = 0#
    MIN = INFINY
    For I = 1 To Nn
        X = pt(I)
        If (X > MAX) Then MAX = X
        If (X <> 0# And X < MIN) Then MIN = X
    Next I
'
'   SCALE ONLY IF THERE ARE VERY LARGE OR VERY SMALL COMPONENTS.
    SFACTOR = 1#
    If (MIN >= LO And MAX <= hi) Then Exit Function
    X = LO / MIN
    If (X > 1#) Then GoTo 20
    SC = 1# / (Sqr(MAX) * Sqr(MIN))
    GoTo 30
'
20
    SC = X
    If (INFINY / SC > MAX) Then SC = 1#
'
30
    L = Log(SC) / Log(BASE) + 0.5
    SFACTOR = BASE ^ L
'
'
'
End Function

Private Sub CDIVID(ByVal ar#, ByVal ai#, ByVal br#, ByVal bi#, cr#, ci#)
'
'   COMPLEX DIVISION C = A/B, AVOIDING OVERFLOW.
'
    Dim d#, r#, t#
'
    If (br <> 0# Or bi <> 0#) Then GoTo 10
'
'   DIVISION BY ZERO, C = INFINY.
    Call MCON(t, INFINY, t, t)
    cr = INFINY
    ci = INFINY
    Exit Sub
'
10
    If (Abs(br) >= Abs(bi)) Then GoTo 20
    r = br / bi
    d = bi + r * br
    cr = (ar * r + ai) / d
    ci = (ai * r - ar) / d
    Exit Sub
'
20
    r = bi / br
    d = br + r * bi
    cr = (ar + ai * r) / d
    ci = (ai - ar * r) / d
'
'
'
End Sub

Private Function CMOD(ByVal re#, ByVal im#) As Double
'
'   MODULUS OF A COMPLEX NUMBER AVOIDING OVERFLOW.
'
    Dim ar#, ai#
'
    ar = Abs(re)
    ai = Abs(im)
    If (ar >= ai) Then GoTo 10
    CMOD = ai * Sqr(1# + (ar / ai) ^ 2)
    Exit Function
'
10
    If (ar <= ai) Then GoTo 20
    CMOD = ar * Sqr(1# + (ai / ar) ^ 2)
    Exit Function
'
20
    CMOD = ar * Sqr(2#)
'
'
'
End Function

Private Sub MCON(ByRef vbETA#, ByRef vbINFINY#, ByRef vbSMALNO#, ByRef vbBASE#)
'
'   Non usata.
'   ----------
'
'   MCON PROVIDES MACHINE CONSTANTS USED IN VARIOUS PARTS OF THE
'   PROGRAM. THE USER MAY EITHER SET THEM DIRECTLY OR USE THE
'   STATEMENTS BELOW TO COMPUTE THEM. THE MEANING OF THE FOUR
'   CONSTANTS ARE:
'   ETA:    THE MAXIMUM RELATIVE REPRESENTATION ERROR
'           WHICH CAN BE DESCRIBED AS THE SMALLEST POSITIVE
'           FLOATING-POINT NUMBER SUCH THAT 1.0D0 + ETA IS
'           GREATER THAN 1.0D0.
'   INFINY: THE LARGEST FLOATING-POINT NUMBER
'   SMALNO: THE SMALLEST POSITIVE FLOATING-POINT NUMBER
'   BASE:   THE BASE OF THE FLOATING-POINT NUMBER SYSTEM USED
'
'   LET T BE THE NUMBER OF BASE-DIGITS IN EACH FLOATING-POINT
'   NUMBER(DOUBLE PRECISION). THEN ETA IS EITHER .5*B**(1-T)
'   OR B**(1-T) DEPENDING ON WHETHER ROUNDING OR TRUNCATION
'   IS USED.
'   LET M BE THE LARGEST EXPONENT AND N THE SMALLEST EXPONENT
'   IN THE NUMBER SYSTEM. THEN INFINY IS (1-BASE**(-T))*BASE**M
'   AND SMALNO IS BASE**N.
'
    Dim vbM&, vbN&, vbT&
'
    ' Visual Basic 6.0 Double data type:
    ' vbN = -323: vbM = 308: BASE# = 10: vbT = 15
    ' INFINY = (1# - 10 ^ (-15)) * 10 ^ 308 = 9.99999999999999E+307
    ' SMALNO = 10 ^ -323 = 9.88131291682493E-324
    ' ETA = 10 ^ (1 - 15) = 0.00000000000001
    ' 1# + ETA = 1# + 10 ^ (1 - 15) = 1.00000000000001
'
    vbBASE = 10#
    vbT = 15
    vbM = 308
    vbN = -323
    vbETA = vbBASE ^ (1 - vbT)
    vbINFINY = vbBASE * (1# - vbBASE ^ (-vbT)) * vbBASE ^ (vbM - 1)
    vbSMALNO = (vbBASE ^ (vbN + 3)) / vbBASE ^ 3
'
'
'
End Sub

